/* main.cpp : Main file for SIP Conference Server.
 * (c) 1999-2001 Columbia University. All rights reserved.
*/
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>        /* strftime() */
extern "C" {
#include <pthread.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
}

#ifndef USE_MYSQL
#error "sipconf needs mysql database to run. Please configure --with-mysql option turned on. "
#endif

#include "rtpmapi.h"
#include "iptelapi.h"
#include "sipapi.h"

extern "C" {
#include "raise-rlimits.h" 
#include "uri.h"
#include "config.h"
#include "conf-util.h"     /* GetDBURI */
#include "dbapi.h"
#include "strdupn.h"
#include "pthread-util.h"
extern void do_daemon(void);
extern char* optarg;
}

/*
 * Define this macro for workaround of plugnsip specific bugs
 */
#define INTEROPERATE_WITH_PLUGNSIP 1 // wait for sometime before accepting
#define SESSION_PRINT  1  //debug messages

/**
 * MySIPUserAgent and MySIPCall
 */
class MySIPUserAgent : public SIPEndpoint
{
public:
  MySIPUserAgent(bool IsServer = true, bool IsRegistrar=false);
  MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar=false);
  virtual IptelCall *AllocateNewCallObject(const char *callid);
};

class MySIPCall : public SIPCall
{
private:
  int rtp; /* RTPMixer context for the call leg */
  char * CanAccept(); 
  session_description_t GetRemoteSDP();
  void SetSelfSDP(session_description_t );

  /* Interface add to support load balancing */
  char   *RunScheduleAlgorithm();
  char   *GetConfInstanceServer();
  char   *GetLeastLoadServer();
  char   *GetServerHasCapacity();
  int    GetEventID(int event_group);
  int    GetEventGroupID();
  void   CreateConfInstance(unsigned int conference_id,
			    char *start_time,
			    char *end_time,
			    char *host_instance,
			    char *recording);

public:
  MySIPCall(const char *callid=NULL) : SIPCall(callid) { };
  MySIPCall(SIPEndpoint & ep, const char *callid=NULL) : 
    SIPCall(ep, callid) { };
  virtual ~MySIPCall() { };
  virtual void   OnNewIncomingCall(IptelEndpoint & ep);
  virtual int    OnHangup();

  static char   *GetCurrentTime();
};


/* Dummy constructor */
MySIPUserAgent::MySIPUserAgent(bool IsServer,bool IsRegistrar)
: SIPEndpoint(IsServer,IsRegistrar)
{
}

MySIPUserAgent::MySIPUserAgent(const IptelIPAddress & ip,bool IsRegistrar)
: SIPEndpoint(ip,IsRegistrar)
{
  /* Do nothing here */
}


/* Allocate a new call object of type MySIPCall */
IptelCall *MySIPUserAgent::AllocateNewCallObject(const char *callid)
{
  return new MySIPCall(*this, callid);
}

/*
 * Read the conference list and findout if the
 * call from the particular sender to a particular
 * receiver (destination or conference URI) can
 * be accepted or not.
 * Returns the conference name on success and
 * NULL on failure
 */
static bool acceptall = false;

char * MySIPCall::CanAccept()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
 
  if ( acceptall ) {
    char *n;
    n = dest.GetUser(); /* Accept all conference */
    if ( n == NULL ) {
      n = dest.GetHost();
    }
    return n;
  }

  /*
   * Check if the desctination address is a valid conference
   * identifier.
   */
    
  /* Check if the name implies adhoc conference. *.adhoc@... */
  char *user = dest.GetUser();
  if ( user != NULL ) {
    char *f = strstr(user, ".adhoc");
    if ( f && !strcmp(f, ".adhoc") ) {
      /* Match found in the username */
      /* Do not free user or f. */
      return user;
    }
  }
  else {
    cout << "Can not accept conference with NULL user in URL" <<
      endl;
    return NULL;
  }
  
  /* Try to locate it in the conferences table */
  if ( !isConnectionEstablished() ) {
    /* No connection with mysql server; fail */
    free(user);
    return NULL;
  }
  
  int ret = 0;
  ResultTable table = query("conferences", "url", user, "url", &ret); 
  freeResultTable(table, ret, 1);
  
  if ( ret <= 0 ) {
    free(user);
    return NULL;
  }
  
  /* If ret > 0 then it is found. return the name */
  return user;
}

/* 
 * Return the current time as a string of format "YYYY-MM-DD hh:mm:ss".
 */
char *MySIPCall::GetCurrentTime()
{
  char *current_time = (char *) malloc(30);
  char *format = "%Y-%m-%d %H:%M:%S";
  size_t size = 0;
  time_t clock;
  struct timeval tv;
  struct timezone tz;

  if(gettimeofday(&tv, &tz) < 0) {
    cout << "Error in getting current time " << endl;
    free(current_time);
    exit(1);
  }

  clock = (time_t) tv.tv_sec;  
  struct tm *timeptr = localtime(&clock);
  
  if ((size = strftime(current_time, 30, format, timeptr)) < 0) {
    cout << "Error in getting current time" << endl;
    free(timeptr);
    free(current_time);
    exit(1);
  }
       
  free(timeptr);

  /* Return the current time */
  return current_time;
}

/* 
 * Return the server name which host the conference if the instance 
 * already exist, Return NULL if the instance doesn't exist.
 */
char *MySIPCall::GetConfInstanceServer()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *servername = NULL;
  char *confname   = NULL;
  char *current_time = NULL;
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  ResultTable query_result;    /* sql query results */
  int rows = 0;                /* number for rows returned by sql query */
  
  /* Query confinstances table to find server for the conference */
  confname = dest.GetUser();
  current_time = GetCurrentTime();    /* remember to free it */

  //cout << "GetConfInstanceServer: confname=" << confname << " currentTime=" 
  //     << current_time << endl;

  /* Find the conference server from the database */
  sprintf(sqlbuf, 
     "SELECT i.host_instance, i.start_time, i.end_time " 
     "FROM   conferences c, calendar calen, confinstances i "
     "WHERE  c.url = '%s' "
     "AND    c.event_group = calen.event_group_id "
     "AND    calen.event_id = i.event_id " 
     "AND    '%s' >= i.start_time "
     "AND    '%s' <= i.end_time ", confname, current_time, current_time);

  query_result = rawQuery(sqlbuf, &rows); 

  if (!query_result || rows <= 0) {
    free(confname);
    free(current_time);
    return NULL;
  }
  
  /* 
   * If current time is between the start_time and end_time, then return 
   * the server name.
   */
  if (rows > 1) {
    cout << "Rejecting" << endl;
    Reject(500, "Error reading conference instance server");
    free(confname);
    free(current_time);
    return NULL;
  }
  
  /* OK, get the server name */
  servername = query_result[0][0];
  //cout << "OK, find server: servername=" << servername << endl;

  free(confname);
  free(current_time);

  return servername;
}

/*
 * Get the event_id for the conference:
 *
 * Return the event_id in the given event_group that currentTime is in 
 * between the start_time and end_time, return 0 if no match found.
 */
int MySIPCall::GetEventID(int event_group_id)
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *confname   = NULL;
  char *current_time = NULL;
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  ResultTable query_result;    /* general sql result set */ 
  int event_id = 0;         
  int rows = 0;     

  /* Check if the name implies adhoc conference. *.adhoc@... */
  confname = dest.GetUser();

  //cout << "GetEventID: confname=" << confname << " event_group=" 
  //     << event_group_id << endl;

  /* Get event which currentTime is in between start and end time */
  current_time = GetCurrentTime();

  sprintf(sqlbuf,
     "SELECT event_id,start_time,end_time "
     "FROM   calendar WHERE event_group_id = '%d' "
     "AND    '%s' >= start_time AND '%s' <= end_time ", 
     event_group_id, current_time, current_time);
       
  query_result = rawQuery(sqlbuf, &rows);

  if (rows > 1) {
    cout << "Rejecting" << endl;
    Reject(500, "Error reading event_id in calendar");
  }
  else if (rows == 1) {
    /* Get the event_id if found match */
    event_id = atoi(query_result[0][0]);
  }

  free(confname);
  free(current_time);

  return event_id;
}

/*
 * Return the event_group_id for the conference:
 *
 * Search in the conferences table, return event_group if find match 
 * in the conference table, otherwise return 0.
 */
int MySIPCall::GetEventGroupID()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *confname   = NULL;
  ResultTable table;
  int event_group_id = 0;
  int ret  = 0;

  /* Get the conference name */
  confname = dest.GetUser();
  table = query("conferences", "url", confname, "event_group", &ret);
  
  if (ret > 1) {
    cout << "Rejecting" << endl;
    Reject(500, "Error reading event_group_id in database");
  }
  else if (ret == 1) {
    /* Get the event_group_id if found match in database */
    event_group_id = atoi(table[0][0]);
  }

  free(confname);
  freeResultTable(table, ret, 1);  
  
  return event_group_id;
}

/*
 * Get the least loaded server at current time.
 */
char *MySIPCall::GetLeastLoadServer()
{
  char *servername = NULL;     /* least loaded server name */
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  char hostname[256];
  ResultTable query_result;    /* sql result for loaded server */
  int rows = 0;                

  /* Get the current load for all the conference servers */
  sprintf(sqlbuf,
     "SELECT s.server_name, sum(c.current_participant_count) "
     "FROM   confservers s "
     "       LEFT JOIN confinstances i ON s.server_name = i.host_instance " 
     "       LEFT JOIN calendar cal ON i.event_id = cal.event_id " 
     "       LEFT JOIN event_group g ON cal.event_group_id = g.event_group_id "
     "       LEFT JOIN conferences c ON g.event_group_id = c.event_group "
     "GROUP  BY s.server_name");

  query_result = rawQuery(sqlbuf, &rows);

  if (!query_result || rows <= 0) {
    cout << "Rejecting" << endl;
    Reject(503, "No available conference server found");
    return NULL;
  }

  /* Get the least loaded server in the system */
  int index = 0;
  int leastload = atoi(query_result[0][1]);
    
  for (int i = 0; i < rows; i++) {
    int load = atoi(query_result[i][1]);
    //char *server = query_result[i][0];
    
    //cout << "i=" << (i + 1) << " server=" << server << " load=" 
    //	   << load << endl;
      
    if (load < leastload) {
      index = i;
      leastload = load;
    }
  }
  
  servername = query_result[index][0];
  
  /* 
   * Check whether the local server has the same load as the least loaded
   * server, then we don't need to redirect the call to another server 
   * which has the same load as the local server. 
   *
   * The local server should have priority to host the conference if other 
   * servers has the same load.
   */ 

  /* Get the local conference server name */
  if (gethostname(hostname, 255) < 0) {
    cout << "Failed to get hostname." << endl;
    exit(1);
  }

  int flag = 0;

  for (int i = 0; i < rows; i++) {
    int load = atoi(query_result[i][1]);
    char *server = query_result[i][0];
      
    if (load == leastload && !strcmp(server, hostname)) {
      flag = 1;      /* Find match */
      break;
    }    
  }  

  /* If found match, reset the server to the local conference server */
  if (flag == 1) {
    servername = (char *) malloc((strlen(hostname) + 1));
    strcpy(servername, hostname);
  }

  return servername;
}

/*
 * Return the conference server which still have extra capacity to
 * host the new conference.
 */
char *MySIPCall::GetServerHasCapacity()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *confname   = NULL;     /* conference name */
  char *servername = NULL;     /* conference server which has capacity */
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  ResultTable query_result;    /* sql query result table */
  int bandwidth_requirement = 0;/* bandwidth requirement for conference */
  int rows = 0; 

  /* Get the conference name */
  confname = dest.GetUser();

  /* Compute the bandwidth requirement for the conference */
  sprintf(sqlbuf, 
     "SELECT max_participant_count FROM conferences "
     "WHERE  url = '%s' ", confname);     
  query_result = rawQuery(sqlbuf, &rows); 
  
  if (rows != 1) {
    cout << "Rejecting" << endl;
    Reject(500, "Error in find conference bandwidth");
    free(confname);
    return NULL;
  }

  bandwidth_requirement = atoi(query_result[0][0]);

  cout << "GetServerHasCapacity: conf=" << confname << " bandwidth="
       << bandwidth_requirement << endl;

  /* Get the current load for all the conference servers */
  sprintf(sqlbuf,
     "SELECT s.server_name, s.capacity - sum(c.current_participant_count) "
     "                                                   AS left_capacity "
     "FROM   confservers s "
     "       LEFT JOIN confinstances i ON s.server_name = i.host_instance " 
     "       LEFT JOIN calendar cal ON i.event_id = cal.event_id " 
     "       LEFT JOIN event_group g ON cal.event_group_id = g.event_group_id "
     "       LEFT JOIN conferences c ON g.event_group_id = c.event_group "
     "GROUP  BY s.server_name ");

  query_result = rawQuery(sqlbuf, &rows);

  if (!query_result || rows <= 0) {
    cout << "Rejecting" << endl;
    Reject(503, "No available conference server found");
    free(confname);
    return NULL;
  }
    
  /* Get the conference server which has the most free capacity */
  int index = 0;
  int max_capacity = atoi(query_result[0][1]);

  for (int i = 0; i < rows; i++) {
    //char *server = query_result[i][0];
    int left_capacity = atoi(query_result[i][1]);

    //cout << "i=" << (i + 1) << " server=" << server << " left_capacity=" 
    //	   << left_capacity << endl;

    if (left_capacity > max_capacity) {
      index = i;
      max_capacity = left_capacity;
    }	
  }

  /* 
   * If the maximum left capacity > conference bandwidth requirement,
   * return the servername, otherwise return NULL.
   */

  if (max_capacity >= bandwidth_requirement) {
    servername = query_result[index][0];
  }
  
  return servername;
}

/*
 * Create the conferenc instance in the database.
 */
void MySIPCall::CreateConfInstance(unsigned int event_id,
				   char *start_time,
				   char *end_time,
				   char *host_instance,
				   char *recording)
{
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  ResultTable query_result;    /* sql query result table */
  int rows = 0;  

  //cout << "CreateConfInstance: event=" << event_id << " start_time=" 
  //     << start_time << " end_time=" << end_time << " host=" 
  //     << host_instance << endl;

  /* In case recording is NULL */
  if (recording == NULL) {
    recording = "";
  }

  /* Run the SQL query to add the conference instance */
  sprintf(sqlbuf,
     "INSERT INTO confinstances (instance_id,event_id, "
     "       start_time,end_time,host_instance,recording) "
     "VALUES(NULL,'%d','%s','%s','%s','%s') ", 
     event_id, start_time, end_time, host_instance, recording);

  query_result = rawQuery(sqlbuf, &rows);
}

/*
 * Run the scheduling algorithm to find the best the server to host
 * the conference.
 *
 * Note that conferences can be created in three ways:
 *
 * 1) "ad hoc" via INVITE to a new conference name that doesn't exist 
 *    yet. In that case, we don't know the conference size and should 
 *    just pick the least loaded server.
 *
 * 2) via the web page for immediate beginning; similar to case (1), 
 *    except that we know the resource need.
 *
 * 3) via the web page to start at some time in the future, possibly 
 *    with repeats. Run the scheduling algorithm to find a server that 
 *    has available capacity for that timeslot and create conference 
 *    instances.
 */
char *MySIPCall::RunScheduleAlgorithm()
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *servername = NULL;
  char *confname   = NULL;
  char *username   = NULL;
  char *current_time = NULL;
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  char strbuf[100];            /* temporary string buffer */
  ResultTable query_result;    /* general sql result set */ 
  ResultTable table;           /* simple sql result set */
  int DEFAULT_ADHOC_COUNT = 10;
  int event_group_id = 0;      /* event group id for the conference */
  int event_id = 0;            /* event id for the conference instance */
  int rows = 0;     
  int ret  = 0;

  /* Check if the name implies adhoc conference. *.adhoc@... */
  confname = dest.GetUser();

  if ( confname != NULL ) {
    char *f = strstr(confname, ".adhoc");

    if ( f && !strcmp(f, ".adhoc") ) {
      /* 
       * Case 1: "ad hoc" via INVITE to a new conference name that doesn't 
       *         exist yet. In that case, we don't know the conference size 
       *         and should just pick the least loaded server.
       */
      cout << "Case 1: .adhoc - add conference to the database" << endl;

      /* Get the username for the participant */
      username = src.GetURI();   /* remember free username */
      current_time = GetCurrentTime();

      if (username == NULL) {
        cout << "Rejecting -- src.GetURI returned null" << endl;
        Reject(400, "Bad Request: Source URI was null");
        free(confname);
        free(current_time);
        return NULL;
      }
      sprintf(strbuf,"%s", username);
      strcpy(username,strbuf + 4);    /* ignore the sip: prefix */
  
      /* 
       * Add the .adhoc conference to the conference table if there 
       * is no record in the database yet 
       */
      table = query("conferences", "url", confname, "url", &ret); 
      freeResultTable(table, ret, 1);

      /* 
       * If ret <= 0 then the conference is not found, we should:
       * 
       *    1.  create the event group in the event_group table
       *    2.  create the conference in the conferences table
       */            
      if ( ret <= 0 ) {
        /* Create the event_group for the new conference */
        sprintf(sqlbuf,
           "INSERT INTO event_group (event_group_id,recurrence,creator,"
           "       group_description,dtstart,dtend) "
           "VALUES(NULL,' ','%s','.adhoc conference','%s', "
           "       '9999-12-31 00:00:00') ", username, current_time);

        query_result = rawQuery(sqlbuf, &rows);

        /* Get the event_group_id back */
        query_result = rawQuery("SELECT LAST_INSERT_ID()", &rows);

        if (rows != 1) {
          cout << "Rejecting" << endl;
          Reject(500, "Error in reading last inserted event_group_id");
          free(confname);
          free(username);
          free(current_time);
          return NULL;	  
        }

        event_group_id = atoi(query_result[0][0]);
        //cout << "event_group_id = " << event_group_id << endl;

        /* Create the conference in database */
        sprintf(sqlbuf,
           "INSERT INTO conferences (url,description,start_time,end_time, "
           "       authentication,conference_type,participant_list_type, "
           "       media_allowed,max_participant_count,password, "
           "       current_participant_count,owner,id,recordingformat, "
           "       recording_urls,event_group) "
           "VALUES('%s',NULL,NULL,NULL,'none','public','public', "
           "       'audio video','%d',NULL,'0','%s',NULL,NULL,NULL,'%d') ", 
           confname, DEFAULT_ADHOC_COUNT, username, event_group_id);

        query_result = rawQuery(sqlbuf, &rows);
      }

      /* 
       * Now the event_group and conference already exist, we should:
       *
       *    1.  create the event in the calendar table
       *    2.  create the conference instance in confinstances table
       */

      event_group_id = GetEventGroupID();

      /* Reject the call if can't find the event_group_id */
      if (event_group_id == 0) {
	cout << "Rejecting" << endl;
	Reject(500, "Error in reading event_group_id");
	free(confname);
	free(username);
	free(current_time);
	return NULL;	
      }

      /* Go ahead to create the event in the calendar */
      sprintf(sqlbuf,
         "INSERT INTO calendar (event_id,event_group_id,start_time, "
         "       end_time,access,description) "
         "VALUES(NULL,'%d','%s','9999-12-31 00:00:00','public',NULL) ", 
         event_group_id, current_time);

      query_result = rawQuery(sqlbuf, &rows);

      /* Get the event_id back */
      query_result = rawQuery("SELECT LAST_INSERT_ID()", &rows);

      if (rows != 1) {
	cout << "Rejecting" << endl;
	Reject(500, "Error in reading last inserted event_id ");
	free(confname);
	free(username);
	free(current_time);
	return NULL;	  
      }
      
      event_id = atoi(query_result[0][0]);

      //cout << "event_group_id=" << event_group_id << "  event_id=" 
      //     << event_id << endl;

      /* Get the current least loaded conference server */
      servername = GetLeastLoadServer();

      if (servername == NULL) {
	cout << "Rejecting" << endl;
	Reject(503, "Error in find least loaded server");
	free(confname);
	free(username);
	free(current_time);
	return NULL;
      }

      cout << "Find least loaded server = " << servername << endl;
      char *end_time = "9999-12-31 00:00:00";

      CreateConfInstance(event_id, current_time, end_time, servername, NULL);
    }
    else {
      /*
       * Case 2+3: If currentTime is between the start_time and end_time 
       *           of any server instance in the calendar table, then:
       *
       *           1. find a server that has available capacity,
       *           2. create the conference instance
       */
      cout << "Case 2+3: Get the scheduled event " << endl;

      ResultTable table = query("conferences", "url", confname, 
				"event_group", &ret);

      /* If the conference doesn't exist, reject the call */
      if ( ret <= 0 ) {
	cout << "Rejecting" << endl;
	Reject(604, "Conference with this URL does not exist");
	freeResultTable(table, ret, 1);
	free(confname);
	return NULL;
      } 
      else {
	freeResultTable(table, ret, 1);
	event_group_id = atoi(table[0][0]);
      }

      /* If not within the scheduled time slot, reject the call */
      event_id = GetEventID(event_group_id);

      if (event_id == 0) {
	cout << "Rejecting" << endl;
	Reject(603, "Not in the conference scheduled time");
	free(confname);
	return NULL;
      }

      //cout << "event_group_id=" << event_group_id << "  event_id=" 
      //     << event_id << endl;

      /* Get the server which still has capacity */
      servername = GetServerHasCapacity();

      if (servername == NULL) {
	cout << "Rejecting" << endl;
	Reject(603, "No available server has enough capacity");
	free(confname);
	return NULL;
      }

      cout << "Find server has capacity = " << servername << endl;

      /* Get the time to create the conference instance */
      sprintf(sqlbuf, 
         "SELECT start_time,end_time FROM calendar "
         "WHERE  event_id = '%d' ", event_id);

      query_result = rawQuery(sqlbuf, &rows);

      if (!query_result || rows == 0 ) {
	cout << "Rejecting" << endl;
	Reject(500, "Error in reading calendar table");
	free(confname);
	return NULL;
      }

      current_time = GetCurrentTime();
      char *end_time = query_result[0][1];

      CreateConfInstance(event_id, current_time, end_time, servername, NULL);
    }  

    free(confname);
    free(current_time);

    /* Return the server for the conference instance */
    return servername;
  }    
 
  /* Reject the call for NULL URL */
  cout << "Rejecting" << endl;
  Reject(400, "Bad Request: NULL user in conference URL");
  free(confname);

  return NULL;
}

/*
 * OnNewIncomingCall: This is called from the SIP library
 * when an incoming call is received
 */
void MySIPCall::OnNewIncomingCall(IptelEndpoint & ep)
{
  IptelAddress dest    = GetRequestURI();
  IptelAddress src     = GetRemoteAddress();
  char *confname = NULL;
  char *username = NULL;
  char *conftype = NULL;       /* private or restricted */ 
  char *servername = NULL;     /* conference server name */
  char hostname[256];          /* conference server host */
  char sqlbuf[2056];           /* sql buffer for constructing queries */
  char strbuf[100];            /* temporary string buffer */
  ResultTable query_result;    /* sql query results */
  int rows = 0;                /* number for rows returned by sql query */
  int current_participant_count = 0;
  int max_participant_count = 0;
  int restricted = 0;          /* conference retricted flag */
  int wildcard_match = 0;      /* username match based on wildcard */

  cout << src << " wants to join " << dest << endl;

  /*
   *  XXX extract the database table for this conference.
   * If the conference URL is not found then return 604.
   * If the conference type is public then skip this step.
   * Else, match the src address agains the participant
   * list for the conference. If matches then continue,
   * otherwise return 400 Forbidden.
   * Check the current_participant_count against
   * max_participant_count and return 500 Conference Full
   * if the limit is reached.
   */
 
  if ( (confname = CanAccept()) == NULL ) {
    /* Can not accept this call leg */
    cout << "Rejecting" << endl;
    Reject(604, "Conference with this URL does not exist");
    return;
  }

  cout << "CanAccept() successful" << endl;

  /* 
   * Add by Huwei Zhang:
   *
   * Get the hostname for running the server 
   */
  if (gethostname(hostname, 255) < 0) {
    cout << "Failed to get hostname." << endl;
    exit(1);
  }

  servername = GetConfInstanceServer();

  if (servername != NULL) {
    /*
     * If the conference instance is on this server, go on to check
     * the user permission.
     * 
     * Otherwise, redirect the user to the server which host the 
     * conference.
     */

    cout << "Conference exists, confserver=" << servername << endl;

    /* Redirect the user to the server which host the conference */
    if (strcmp(servername, hostname)) {
      IptelAddress newdest(dest);
      newdest.SetHost(servername);

      int result = Redirect(newdest, "not the conference host");
      if (result < 0) {
	cout << "Rejecting: Redirect to server: " << servername 
	     << " failed" << endl;
	Reject(500, "Redirect to another server failed");
      } else {
	cout << "Redirect to server: " << servername << " succeed" << endl;
      }

      return;
    }

    cout << "Local server will host the conference " << confname << endl;
  } 
  else {    
    /* 
     * Run the scheduling algorithm to find the best server to host
     * the conference.
     */    
    cout << "New conference: RunScheduleAlgorithm()" << endl;

    servername = RunScheduleAlgorithm();
    if (servername == NULL) {
      cout << "Rejecting" << endl;
      Reject(500, "No available server found for the conference"); 
      return;        /* Can not accept this call leg */
    }

    cout << "RunScheduleAlgorithm: server=" << servername << endl;
    
    /* Redirect the user to the server which host the new instance */
    if (strcmp(servername, hostname)) {
      IptelAddress newdest(dest);
      newdest.SetHost(servername);

      int result = Redirect(newdest, "not the conference host");
      if (result < 0) {
	cout << "Rejecting: Redirect to server: " << servername 
	     << " failed" << endl;
	Reject(500, "Redirect to another server failed");
      } else {
	cout << "Redirect to server: " << servername << " succeed" << endl;
      }

      return;
    }

    cout << "Local server will host the conference " << confname << endl;
  }

  /* Go on to user authentication and update database */
  sprintf(sqlbuf,"select current_participant_count,max_participant_count,"
	  "conference_type from conferences where url='%s' ",confname);

  query_result = rawQuery(sqlbuf, &rows); 
 
  if (!query_result || rows <= 0) {
    //fatal("OnNewIncomingCall","Eror reading current/max_participant_count");
    Reject(500, "Error reading participant count from database");
    return;
  }

  /* the following condition shouldn't happen, but checking anyway */
  if (rows != 1) {
    //fatal("OnNewIncomingCall","Eror reading current/max_participant_count." 
    //                                       "Too many rows returned");
    Reject(500, "Error reading participant count from database");
    return;
  }

  current_participant_count = atoi(query_result[0][0]);
  max_participant_count = atoi(query_result[0][1]);
  conftype = strdup(query_result[0][2]);
  if (!strcmp(conftype,"restricted")) {
    restricted = 1;       /* conference is restricted */
  }
  printf("current = %d, max = %d restricted = %d\n", current_participant_count,
	 max_participant_count, restricted);

  /* reject call if maximum participant count is already reached */ 
  if (current_participant_count >= max_participant_count) {
    cout << "Rejecting -- max count" << endl;
    Reject(500, "Maximum Participant Count Reached");
    free(confname);
    return;
  }

  /* find out the username of the caller */
  //username = (char *) calloc(80, sizeof(char));
  username = src.GetURI();   // remember free username
  if (username == NULL) {
    cout << "Rejecting -- src.GetURI returned null" << endl;
    Reject(500, "Source URI was null");
    free(confname);
    return;
  }
  sprintf(strbuf,"%s", username);
  strcpy(username,strbuf + 4);    /* ignore the sip: prefix */
  cout << "username = " << username << " conf name = " << confname << endl;

  /*
   * XXX Check if authentication is required; if yes
   * call the appropriate method. Also, be sure
   * to use the user:passwd as url:password or
   * src:password;
   */

  if (restricted)
  {
    /* check to see if the user is in the confusers table */
    /* first check for user name, then "*@user's_domain", finally for "*" */

    /* select dummy field (status) to see it user matching confurl exists */
    sprintf(sqlbuf,"select status from confusers where confurl='%s' "
                   "and user ='%s';",confname, username);
    query_result = rawQuery(sqlbuf, &rows); 

    if (rows < 0) {
      //fatal("OnNewIncomingCall","Eror reading confusers table");
      Reject(500,"Error reading database");
      return;
    }
      
    if (rows == 0) {
      /* now check whether "@callers_domain" exists in confusers table */

      char *domainname; /* caller's domain name */
      domainname =  strchr(strbuf,'@'); 
      wildcard_match = 1;

      
      if (username == NULL) {
        cout << "Rejecting -- malformed username" << endl;
        Reject(500, "Malformed source URI");
        free(confname);
        return;
      }

      cout << "domainname = " << domainname << endl;
      sprintf(sqlbuf,"select status from confusers where confurl='%s' "
	             "and user ='%s';",confname, domainname);
      query_result = rawQuery(sqlbuf, &rows); 

      if (rows < 0) {
        //fatal("OnNewIncomingCall","Eror reading confusers table");
        Reject(500,"Error reading database");
        return;
      }

      if (rows == 0) {
	/* now check whether "*" exists in confusers table */
      
        sprintf(sqlbuf,"select status from confusers where confurl='%s' "
		       "and user ='*';",confname);
        query_result = rawQuery(sqlbuf, &rows); 

        if (rows < 0) {
          //fatal("OnNewIncomingCall","Eror reading confusers table");
          Reject(500,"Error reading database");
          return;
        }

        if (rows == 0) {
          cout << "Rejecting -- conference is restricted" << endl;
          Reject(603,"Conference is restricted");
          free(confname);
          return;
        }
      }
    }
   
    
    bool authRet;

    /* if the user is in the confusers table, authenticate */
    authRet = Authenticate("digest", "non-existant");
    if (authRet == true) {
      cout << "Authenticate returned true" << endl;
    }
    else {
      cout << "Authenticate returned false" << endl;
      Reject();
      free(confname);
      return;
    }
  } /* if restricted */


  /*
   * XXX check media allowed field and format
   * the SDP for the conference based on that.
   */

  /* Accept the call */
  /* Set the session description first */
  session_description_t sdp;
  sdp = GetRemoteSDP();

#ifdef SESSION_PRINT
  cout << "Participant's session:" << endl;
  SDPPrint(&sdp);
  cout << flush;
#endif

  /* Add the call leg in the mixer library */
  rtp = RTPMAddCallLeg(confname /* name of conference */,
                       src.GetAddressString());

  if ( rtp < 0 ) {
    SDPFree(&sdp);
    cout << "Rejecting due to internal failure" << endl;
    Reject(500, "Could not create mixer call leg"); 
    free(confname);
    /* could not create the RTP Mixer leg */
    return;
  }

  /* Set the self SDP of mixer */
  if ( RTPMSetSelfSessionDescription(rtp, sdp) < 0) {
    SDPFree(&sdp);
    RTPMRemoveCallLeg(rtp);
    Reject(500, "Error in setting session description");
    free(confname);
    return;
  }
  SDPFree(&sdp);

  /* Get the mixer's SDP */
  sdp = RTPMGetRemoteSessionDescription(rtp);

#ifdef SESSION_PRINT
  cout << "Server's session:" << endl;
  SDPPrint(&sdp);
  cout << flush;
#endif

  SetSelfSDP(sdp);
  SDPFree(&sdp);

#ifdef INTEROPERATE_WITH_PLUGNSIP
  /* XXX what if the call is closed during this time */
  sleep(1);
#endif

  /*
   * XXX update the current participant count in the conferences table.
   */

  ++current_participant_count;

  //sprintf(sqlbuf,"update conferences set current_participant_count = '%s' where url='%s';",itoa(current_participation_count,strbuf,10),confname);

  cout << "updating participant count " << endl; 
  sprintf(strbuf,"%d", current_participant_count);
  const char * column_list[4]; 
  const char * value_list[4]; 
  int retval;
  column_list[0] = "#current_participant_count";
  value_list[0] = strbuf;
  cout << "before query " << endl; 
  retval = updateOnly("conferences", "url", confname, column_list , value_list, 1) ;
  cout << "after query " << endl; 
  if (retval == -1) {
    //fatal("OnNewIncomingCall","Eror updating conferences table");
    Reject(500,"Error updating database");
    return;
  }


  /* XXX
   * Check the conf users' table and update the
   * send, receive, and audio video parameters for the
   * callleg. Update the type to dialin.
   * update the status to connected.
   * Only done if call is restricted
   */

  if (restricted && !wildcard_match) {
    cout << "updating status" << endl; 
    column_list[0] = "type";
    value_list[0] = "dialin";
    column_list[1] = "status";
    value_list[1] = "connected";
    retval = updateOnly2("confusers", "confurl", confname, "user", username, 
			 column_list , value_list, 2, 0) ;

    if (retval == -1) {
      //fatal("OnNewIncomingCall","Error updating call status in confusers table");
      Reject(500,"Error updating database");
      return;
    }
  }

  cout << "freeing pointers " << endl; 
  free(confname);
  //free(username);

  Accept(); /* Accept the call finally */

}

/*
 * A separate thread is used to
 * terminate the call leg in the
 * RTP AUdio mixer library.
 * This is to avoid delay in response to SIP
 * entity
 */
void *RemoveCallLegThread (void *p)
{
  //cout << "RemoveCallLegThread is called" << endl;

  int rtp = (int) p;
  if ( rtp >= 0 ) {
    /*
    if ( debug_flag ) {
      cout << "Calling RTPMRemoveCallLeg from RemoveCallLegThread" <<
        endl;
    }
    */
    RTPMRemoveCallLeg(rtp);
  }
  return NULL;
}

/*
 * This function is called by the 
 * SIP library when the SIP call is
 * closed by remote
 */
int   MySIPCall::OnHangup()
{
  IptelAddress dest    = GetSelfAddress();
  IptelAddress src     = GetRemoteAddress();

  pthread_t tid;
  char sqlbuf[2056];   /* sql buffer for constructing queries */
  char strbuf[40];     /* temporary string buffer */
  char *username = NULL;    /* username of caller */
  char *confname = NULL;    /* name of conference */
  char *conftype = NULL;    /* type of conference */
  ResultTable query_result; /* sql query result */
  int rows = 0;             /* # of rows returned by sql query */
  int current_participant_count = 0;
  int restricted = 0;       /* conference restricted flag */

  /* Create a thread to close RTP Mixer Library call leg */
  if ( pthread_create_detached(&tid, NULL, RemoveCallLegThread, (void *) rtp)
       !=0) {
    cerr << "pthread_create_detached failed in OnHangup" << endl;
    RemoveCallLegThread((void *)rtp);
  }

  /* 
   * Terminate the call leg from this user
   * for the appropriate conference 
   */
  cout << src << " leaving conference " << dest << endl;

  /* XXX
   * Update the confusers table to 
   * reflect, status=notconnected. 
   * update the conferences table to reflect 
   * current_participant_count.
   */

  /* first, get current_participant_count, decrement it, then update table */

  confname = dest.GetUser();
  username = src.GetURI();

  if (confname == NULL) {
    cout << "Conference URL not found" << endl;
    return 0;
  }
  sprintf(sqlbuf,"select current_participant_count, conference_type "
	  "from conferences where url='%s';",confname);
  query_result = rawQuery(sqlbuf, &rows); 
 
  if (!query_result || rows <= 0) {
    error("OnHangUp","Eror reading current_participant_count");
    return 0;
  }

  /* the following condition shouldn't happen, but checking anyway */
  if (rows != 1) {
    error("OnHangUp","Eror reading current_participant_count. Too many rows returned");
    return 0;
  }

  current_participant_count = atoi(query_result[0][0]);
  conftype = strdup(query_result[0][1]);
  if (!strcmp(conftype,"restricted")) {
    restricted = 1;       /* conference is restricted */
  }
  cout << "current participant count = " << current_participant_count 
       << "  restricted = " << restricted << endl;

  --current_participant_count;
 
  sprintf(strbuf,"%d", current_participant_count);
  const char * column_list[4]; 
  const char * value_list[4]; 

  int retval;
  column_list[0] = "#current_participant_count";
  value_list[0] = strbuf;
  retval = updateOnly("conferences", "url", confname, column_list , value_list, 1) ;
  if (retval == -1) {
    error("OnHangUp","Eror updating conferences table");
    return 0;
  }


  /*
   * Add by Huwei Zhang:
   *
   * Here: 
   *     If the conference is ".adhoc" and current_participant_count = 0,
   *     we should:
   *
   *     1. terminate the conference instance, update the end_time
   *     2. terminate the event in calendar, update the end_time
   */

  if (current_participant_count == 0) {    
    char *f = strstr(confname, ".adhoc");

    if ( f && !strcmp(f, ".adhoc") ) {
      /* Match found in the confname */

      /* Get the event_group_id */
      int event_group_id = GetEventGroupID();

      if (event_group_id == 0) {
	cout << "Error in reading event_group_id" << endl;
	return 0;	
      }

      /* Get the event_id */
      int event_id = GetEventID(event_group_id);

      if (event_id == 0) {
	cout << "Error in reading event_id" << endl;
	return 0;	
      }

      /* Update the conference instance table */
      char *current_time = GetCurrentTime();

      sprintf(sqlbuf,
         "UPDATE confinstances SET end_time='%s' "
	 "WHERE  event_id='%d'", current_time, event_id);

      query_result = rawQuery(sqlbuf, &rows);

      /* Update the event in calendar table */
      sprintf(sqlbuf,
         "UPDATE calendar SET end_time='%s' "
	 "WHERE  event_id='%d'", current_time, event_id);

      query_result = rawQuery(sqlbuf, &rows);
    }
  }


  /* Now, update status of caller to notconnected in confusers */

  if (restricted) {
    username = src.GetURI();   
    if (username == NULL) {
      cout << "src.GetURI returned null" << endl;
      return 0;
    }
    sprintf(strbuf,"%s",username);
    strcpy(username,strbuf + 4);
    cout << "username = " << username << " confname = " << confname << endl;
  
    column_list[0] = "type";
    value_list[0] = "dialin";
    column_list[1] = "status";
    value_list[1] = "notconnected";
    retval = updateOnly2("confusers", "confurl", confname, "user", username, 
			 column_list , value_list, 2, 0) ;

    if (retval == -1) {
      error("OnHangup","Eror updating confusers table");
      return 0;
    }
  
    if (retval == 0) {
      cout << "OnHangup: conference, username pair not found in confusers table" << endl;
    }
  } /* if retricted */

  free(confname);

  SIPCall::OnHangup();
  return 0; /* We should use Mutex in OnNewIncomingCall to
               protect againts freeing of call object */
}

/*
 * Convert the C++ object to C structure for
 * session description of remote entity 
 */
session_description_t MySIPCall::GetRemoteSDP()
{
  session_description_t sdp;
  SDPInit(&sdp);

  SessionDescription *session = GetRemoteSessionDescription();
  if ( session ) {
    for ( int i=0; i<session->GetMediaCount(); i++ ) {
      MediaInfo & media = (*session)[i];
      media_line_t m;
      ListCreate(&m.payloads,1,1);
      for ( int j=0; j<media.GetFormatCount(); j++) {
        FormatInfo & format = media[j];
        payload_type_t p;
        char description[100];
        int clockrate = format.GetClockRate();
        int channel_count = format.GetChannelCount();
        const char *name = format.GetFormatName();

        p.pt = format.GetPayloadType();
        cout << "GetSDP: " << p.pt << "=>" << (name != NULL ? name : "NULL")
             << endl;
        if ( name != NULL ) {
          if ( channel_count <= 1 ) {
            sprintf(description, "%s/%d", name, clockrate);
          }
          else {
            sprintf(description, "%s/%d/%d", name, clockrate,
                    channel_count);
          }
          p.description = strdup(description);
        }
        else {
          p.description = NULL;
        }
        ListAddElement(&m.payloads, &p, sizeof(p));
      }

      IptelIPAddress addr;
      addr = media.GetRxAddress();
      addr.GetIpAddress(m.sin);

      m.ttl = 255; /* ignored */
      AFISetPort(&m.sin, media.GetRxPort());
      switch (media.GetDataType() ) {
      case MediaInfo::Audio:
        m.type = RTPM_MEDIA_AUDIO;
        break;
      case MediaInfo::Video:
        m.type = RTPM_MEDIA_VIDEO;
        break;
      case MediaInfo::Application:
        m.type = RTPM_MEDIA_DATA;
        break;
      default:
        m.type = RTPM_MEDIA_UNKNOWN;
        break;
      }

      ListAddElement(&sdp.media_line_t, &m, sizeof(m));
    }
  }

  /* if ( debug_flag ) {
    cout << "Get remote SDP";
    SDPPrint(&sdp);
    cout << endl;
  }
  */
  

  return sdp;
}

/*
 * Convert the C++ object to C structure for
 * session description of the local object 
 */
void MySIPCall::SetSelfSDP(session_description_t sdp)
{
  int i,j;

  SessionDescription *session = new SessionDescription;

  for (i = 0; i < ListCount(&sdp.media_line_t); i++) {
    media_line_t *m = (media_line_t *) 
      ListElement(&sdp.media_line_t, i);
    if ( !m )
      continue;

    MediaInfo media;

    media.SetRxAddress(m->sin);
    media.SetDirection(MediaInfo::Receive);
    switch (m->type ) {
    case RTPM_MEDIA_AUDIO:      
      media.SetDataType(MediaInfo::Audio);
      break;
    case RTPM_MEDIA_VIDEO:      
      media.SetDataType(MediaInfo::Video);
      break;
    case RTPM_MEDIA_DATA:
      media.SetDataType(MediaInfo::Application);
      media.SetTransportType(MediaInfo::UDP_Transport);
      break;
    default:      
      media.SetDataType(MediaInfo::UnknownData);
      break;
    }

    media.SetRxPort(AFIGetPort(&m->sin));

    for ( j=0; j<ListCount(&m->payloads); j++) {
      payload_type_t *p = (payload_type_t *)
        ListElement(&m->payloads, j);
      if ( !p ) 
        continue;
      if ( p->description == NULL ) {
        media.AddFormat(p->pt);
      }
      else {
        char *l = NULL;
        char *description = strdup(p->description);
        char *name = strtok_r(description, "/", &l);
        char *clkstr = strtok_r(NULL, "/", &l);
        char *cnlstr = strtok_r(NULL, "/", &l);
        if ( name == NULL ) {
          media.AddFormat(p->pt);
        }
        else {
          if ( clkstr == NULL ) {
            media.AddFormat(FormatInfo(p->pt, name, 8000) /* deafault */);
          }
          else if ( cnlstr == NULL ) {
            media.AddFormat(FormatInfo(p->pt, name, atoi(clkstr)));
          }
          else {
            media.AddFormat(FormatInfo(p->pt, name, atoi(clkstr),
                            atoi(cnlstr)));
          }
        }
        free(description);
      }
    }

    session->AddSession(media);
  }

  SDPPrint(&sdp);
  cout << "C++ Session : " << *session << endl;

  SetSelfSessionDescription(session);

}

/*
 * XXX new methods to be implemented.
 * Implement OnRemoteSessionChanged() method.
 * Makesure that the permissions for audio,video and
 * privilege is checked for the participant.
 *
 * Implement OnNotify() method to process the
 * commands from web GUI.
 * First authenticate the request with digest.
 * Then extract the XML message body and parse it to
 * the command.
 * Following commands are supported:
 * - New Participant Type for a dialout conference:
 *     Initiate a new Call to the participant. Be prepared to
 *     supply any authentication required.
 * - Participant removed or status changed to notconnected.
 *     Send a Hangup to the participant if he was in
 *     the conference.
 * - Change audio, video, send, receive preference for
 *   a participant.
 *     Update the call leg to reflect the change.
 * - Removal or Termination (end-time) of the conference.
 *     Send Hangup to all the participants.
 */


/*
 * usage()
 * Usage information. Exits program with status code 1.
 */
void usage(char *prog) 
{
  fprintf(stderr,"Usage: %s [options]\n"
  " or    %s -h\n"
  "v        : Print version and exit.\n"
  "h        : Print usage and exit.\n"
  "d        : Use debug (verbose) mode.\n"
  "f file   : Use Database URI configuration file. Default is cinema_db.conf\n"
  "X        : Non-daemon mode.\n"
  "i        : Interactive mode, used for debugging.\n"
  "n        : Use numeric (dotted-decimal) for host name.\n"
  "a        : Accept all conference without looking in conf list.\n"
  "m time   : Packatization time for outgoing RTP packets. Default is 40\n"
  "           for 40 milliseconds\n"
  "p port   : Listening SIP port, default is 5060.\n"
  "b        : Send audio back to the participant who generated it.\n"
  "D database-uri : sql://user:password@db-server/database-name\n"
  "H hostname     : IPv6 host name, if different from normal\n"
	  , prog, prog);
  exit(1);
} /* usage */

/* Main routine */

void ConfigDB() {
  uri_t table_uri;

  /* Old style database configuration */ 
  table_uri = URI_Parse(config.UserConfig);

 if (config.UserConfig == NULL) {
    fatal("ConfigUser", "No UserConfig configured\n");
  }

  /* New-style database configuration */
  config.database = new_cinema_table(config.UserConfig, NULL, NULL);
  if (config.database == NULL) {
    fatal("ConfigUser", "Could not open database");
  }
  
  if (strcasecmp(table_uri.scheme, "sql") == 0) {
    char *password;
    char *path = "sip";
    password = table_uri.password;
    if ( password != NULL && !strcmp(password,"NULL")) {
      password = NULL;
    }
    if (table_uri.path) {
        path = table_uri.path;
    }
    if (initializeConnection(table_uri.user, password, path,
                             table_uri.host, table_uri.port) != 0) {
      fatal("ConfigUser", "Could not open database %s\n", config.UserConfig);
    }
  }
  else {
    fatal("ConfigUser",
          "Unknown database scheme %s in UserConfig pseudo-URL %s\n",
          table_uri.scheme, config.UserConfig);
  }

  URI_Free(&table_uri);
  return;
} /* ConfigDB */
   
int main(int argc, char *argv[])
{
  int c, e;
  bool isdaemon = true;
  bool isinteractive = false;
  int packetization_time = 40;
  unsigned short listening_port = 5060;
  char config_file[256] = { 0 };

  ResultTable query_result;    /* sql query results */
  int rows = 0;
  char hostname[256];          /* conference server host */
  char sqlbuf[2048];           /* sql buffer */

  /* Parse command line options */
  while((c = getopt(argc, argv, "f:vdhXiam:p:nD:bH:")) != EOF) {

    switch(c) {
    case 'v' :
#ifdef VERSION 
      printf("sipconf version %s\n", VERSION);
#else
      printf("sipconf version unknown\n");
      exit(0);
#endif
    case 'h' :
      usage(argv[0]);
      exit(0);
      break;
    case 'd' :
      SIPLibrary::DebugOn("stdout");
      break;
    case 'n' :
      SIPLibrary::oUseNumericAddress = true;
      break;
    case 'f' :
      strncpy(config_file, optarg, sizeof(config_file));
      break;
    case 'X' :
      isdaemon = false;
      break;
    case 'i' :
      isinteractive = true;
      break;
    case 'a' :
      acceptall = true;
      break;
    case 'H' :
      strassign(&SIPLibrary::HostName6, optarg);
      break;
    case 'm':
      packetization_time = atoi(optarg);
      break;
    case 'p':
      listening_port = atoi(optarg);
      break;
    case 'b':
      RTPMSendBackAudioToParticipant(1 /* 1 for true */);
      break;
    case 'D': 
      strassign(&config.UserConfig, optarg);
      break;
    default:
      fprintf(stderr,"Unknown option -%c\n", c);
      usage(argv[0]);
      exit(1); /* Not required */
    }
  }

  startupSocket();
  
  if (config.UserConfig == NULL) {
    config.UserConfig = GetDBURI(config_file);
    
    if (!config.UserConfig) {
      fprintf(stderr, "No database locatable. Use -D or -f \n");
      exit(1);
    }
  }  
  
  ConfigDB();
  
  if ((e = SIPLibrary::CheckLicense(config.UserConfig, 
				    "sipconf_license")) < 0) {
    cerr << "Unlicensed copy. Contact hgs@cs.columbia.edu for a license." << endl;
    //exit(1);
  }
  
  /* Min packetization time is 20 ms, max is 2 seconds */
  if ( packetization_time < 20 ) {
    packetization_time = 20;
  }
  else if ( packetization_time > 2000 ) {
    packetization_time = 2000;
  }

  raise_rlimits();

  /* Create the listening servers */
  IptelIPAddress listeningAddress;
  listeningAddress.SetPort(listening_port);

  SIPLibrary::SetName("SIPCONF-Columbia-University-Audio-Conference-Server");
  if ( isdaemon ) {
    do_daemon();
  }
  else {
    SIPLibrary::ErrorOn("stderr");
  }

  MySIPUserAgent sip(listeningAddress, false);

  /* Set the packetization time */
  RTPMSetPacketizationTime(packetization_time);


  /* 
   * Add by Huwei Zhang:
   *
   * Register the conference server in the database before accept 
   * the connection from user.
   */
  if (gethostname(hostname, 255) < 0) {
    cout << "Failed to get hostname." << endl;
    exit(1);
  }

  sprintf(sqlbuf, "SELECT server_name from confservers " 
                  "WHERE server_name='%s'", hostname);

  query_result = rawQuery(sqlbuf, &rows); 

  /* the following condition shouldn't happen, but checking anyway */
  if (rows > 1) {
    cout << "Error reading confserver from database" << endl;
    exit(1);
  }
  else if (!query_result || rows <= 0) {
    /* Add the server to the confservers table if not exist */ 
    sprintf(sqlbuf, "INSERT INTO confservers (server_name,capacity) "
                    "VALUES('%s', '50')", hostname);    
    
    query_result = rawQuery(sqlbuf, &rows);
  }


  cout << "Waiting for connections..." << endl;

  if (isinteractive) {
    bool oExitFlag = false;
    char command[180];
    
    while ( !oExitFlag ) {
      cout << "sipconf> " << flush;     /* prompt for command */
      strcpy(command,"");
      fscanf(stdin,"%[^\n]",command); /* read command */
      getchar();                      /* remove \n from stdin */
      
      char * cmd = strtok(command," ");
      if ( cmd == NULL ) continue;
      if ( !strcmp(cmd,"exit") || !strcmp(cmd,"quit")) {
        oExitFlag = true;
      }
      else {
        cout << "Command " << command << " not implemented" << endl;
      }
    }
  }
  else {
    while ( 1 ) {
      /* There should be a better way to wait for termination */
#if 1
      sleep(10000);
#else
      struct rusage ubuf;
      sleep(5);
      getrusage(RUSAGE_SELF, &ubuf);
      printf("Memory-maxrss(%ld) idrss(%ld)\n",
             ubuf.ru_maxrss,
             ubuf.ru_idrss);
             
#endif
    }

    cout << "Exiting" << endl;
    exit(0);
  }

  /*
   * close connection to Mysql Database.
   */
  closeConnection();

  free_valid(config.UserConfig);

  cout << "Exiting" << endl;
  return 0; /* Not required */
}

/************************** END OF FILE *********************************/







